from typing import Dict

from google.auth.transport.requests import Request
from google.oauth2 import service_account

from gcp_pilot.base import GoogleCloudPilotAPI


class IdentityAwareProxy(GoogleCloudPilotAPI):
    def __init__(self, audience):
        self.audience = audience
        super().__init__()

    def _build_client(self, **kwargs) -> None:
        return None

    def _get_id_credentials(self):
        credentials = service_account.IDTokenCredentials(
            signer=self.credentials.signer,
            service_account_email=self.credentials.service_account_email,
            token_uri="https://oauth2.googleapis.com/token",  # TODO: fetch from credentials
            target_audience=self.audience,
        )
        if not credentials.valid:
            credentials.refresh(request=Request())
        return credentials

    @property
    def token(self) -> str:
        credentials = self._get_id_credentials()
        return credentials.token

    @property
    def expires_at(self):
        credentials = self._get_id_credentials()
        return credentials.expiry

    @property
    def headers(self) -> Dict[str, str]:
        return {'Authorization': f'Bearer {self.token}'}


__all__ = (
    'IdentityAwareProxy',
)



# More Information: <https://googleapis.dev/python/dns/latest/index.html>
import time
from enum import Enum
from typing import Generator, List

from google.cloud import dns

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI


class RecordType(Enum):
    A = 'A'
    CNAME = 'CNAME'
    MX = 'MX'
    TXT = 'TXT'

    @classmethod
    def prepare(cls, record_type: 'RecordType', record: str):
        if record_type == cls.CNAME:
            return cls.build_dns_name(name=record)
        # TODO Parse MX, A records
        return record

    @classmethod
    def build_dns_name(cls, name: str) -> str:
        if name and not name.endswith('.'):
            dns_name = f'{name}.'
        else:
            dns_name = name
        return dns_name


class CloudDNS(GoogleCloudPilotAPI):
    _client_class = dns.Client

    def _get_client_extra_kwargs(self):
        return {
            'project': self.project_id,
        }

    def list_zones(self) -> Generator[dns.ManagedZone, None, None]:
        yield from self.client.list_zones()

    def _build_zone(self, name: str, dns_name: str, description: str = None) -> dns.ManagedZone:
        return self.client.zone(
            name=name,
            dns_name=RecordType.build_dns_name(name=dns_name),
            description=description,
        )

    def create_zone(
            self,
            name: str,
            dns_name: str,
            description: str = None,
            exists_ok: bool = True,
    ) -> dns.ManagedZone:
        zone = self._build_zone(name=name, dns_name=dns_name, description=description)
        if not zone.exists():
            return zone.create()
        if not exists_ok:
            raise exceptions.AlreadyExists()
        return zone

    def delete_zone(self, name: str, dns_name: str, exists_ok: bool = True) -> None:
        zone = self._build_zone(name=name, dns_name=dns_name)
        if not zone.exists and not exists_ok:
            raise exceptions.NotFound()
        return zone.delete()

    def list_records(self, zone_name: str, zone_dns: str) -> Generator[dns.ResourceRecordSet, None, None]:
        zone = self._build_zone(name=zone_name, dns_name=zone_dns)

        yield from zone.list_resource_record_sets()  # TODO Why paging does not work like docs?

    def _change_record(
            self,
            action: str,
            zone_name: str,
            zone_dns: str,
            name: str,
            record_type: RecordType,
            record_data: List[str] = None,
            ttl: int = 60 * 60,
            wait: bool = True,
    ) -> dns.ResourceRecordSet:
        zone = self._build_zone(name=zone_name, dns_name=zone_dns)

        parsed_record_data = [
            RecordType.prepare(record_type=record_type, record=record) for record in record_data
        ] if record_data else []

        record_set = zone.resource_record_set(
            name=RecordType.build_dns_name(name=name),
            record_type=record_type.name,
            ttl=ttl,
            rrdatas=parsed_record_data,
        )
        changes = zone.changes()

        if action == 'add':
            changes.add_record_set(record_set)
        elif action == 'delete':
            changes.delete_record_set(record_set)
        else:
            raise exceptions.ValidationError(f"Action {action} is not support for record sets")

        changes.create()

        if wait:
            while changes.status != 'done':
                print('Waiting for changes to complete')
                time.sleep(60)
                changes.reload()
        return record_set

    def add_record(
            self,
            zone_name: str,
            zone_dns: str,
            name: str,
            record_type: RecordType,
            record_data: List[str],
            ttl: int = 5 * 60,
            wait: bool = True,
    ) -> dns.ResourceRecordSet:
        return self._change_record(
            action='add',
            zone_name=zone_name,
            zone_dns=zone_dns,
            name=name,
            record_type=record_type,
            record_data=record_data,
            ttl=ttl,
            wait=wait,
        )

    def delete_record(
            self,
            zone_name: str,
            zone_dns: str,
            name: str,
            record_type: RecordType,
            wait: bool = True,
    ) -> dns.ResourceRecordSet:
        return self._change_record(
            action='delete',
            zone_name=zone_name,
            zone_dns=zone_dns,
            name=name,
            record_type=record_type,
            wait=wait,
        )


__all__ = (
    'CloudDNS',
    'RecordType',
)



from contextlib import ExitStack
from unittest.mock import patch
from functools import wraps

from google.oauth2.service_account import Credentials


class patch_auth:
    def __init__(self, project_id: str = 'potato-dev', location: str = 'moon-dark1'):
        # Realistic: actual class to be accepted by clients during validation
        # But fake: with as few attributes as possible, any API call using the credential should fail
        credentials = Credentials(
            service_account_email='chuck@norris.com',
            signer=None,
            token_uri='',
            project_id=project_id,
        )
        managers = [
            patch('google.auth.default', return_value=(credentials, project_id)),
            patch('gcp_pilot.base.GoogleCloudPilotAPI._set_location', return_value=location),
            patch('gcp_pilot.base.AppEngineBasedService._set_location', return_value=location),
        ]
        self.stack = ExitStack()
        for mgr in managers:
            self.stack.enter_context(mgr)

    def __enter__(self):
        return self.stack.__enter__()

    def start(self):
        return self.__enter__()

    def __exit__(self, typ, val, traceback):
        return self.stack.__exit__(typ, val, traceback)

    def stop(self):
        self.__exit__(None, None, None)

    def __call__(self, f):
        @wraps(f)
        def wrapper(*args, **kw):
            with self:
                return f(*args, **kw)
        return wrapper



import json
from dataclasses import field, dataclass
from typing import List, Dict, Generator

import requests

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI, DiscoveryMixin, ResourceType


class Text:
    @classmethod
    def build_mention(cls, member_id: str = 'all') -> str:
        return f"<users/{member_id}>"

    @classmethod
    def build_link(cls, url: str, text: str) -> str:
        return f"<{url}|{text}>"

    @classmethod
    def format_color(cls, hex_color: str, text: str) -> str:
        return f"<font color=\"#{hex_color}\">{text}</font>"


class Widget(dict):
    _key = None

    def as_data(self):
        if self._key:
            return {
                self._key: dict(self)
            }
        else:
            return dict(self)


class ButtonWidget(Widget):
    def __init__(self, url, text: str = None, image_url: str = None, icon: str = None):
        super().__init__()
        if text:
            self._key = 'textButton'
            self['text'] = text
        elif image_url:
            self._key = "imageButton"
            self['iconUrl'] = image_url
        elif icon:
            self._key = 'imageButton'
            self['icon'] = icon
        else:
            raise exceptions.UnsupportedFormatException("A button must have a text, image or icon")

        self["onClick"] = {
            "openLink": {
                "url": url,
            }
        }


class OnClickWidget(Widget):
    _key = 'onClick'

    def __init__(self, url):
        data = dict(
            openLink=dict(url=url)
        )
        super().__init__(data)


class KeyValueWidget(Widget):
    _key = 'keyValue'

    def __init__(
            self,
            content: str,
            top: str = None,
            bottom: str = None,
            break_lines: bool = True,
            on_click: OnClickWidget = None,
            icon: str = None,
            button: ButtonWidget = None,
    ):
        data = dict(
            content=content,
            contentMultiline='true' if break_lines else 'false',
        )
        if top:
            data['topLabel'] = top
        if bottom:
            data['bottomLabel'] = bottom
        if on_click:
            data['onClick'] = on_click
        if icon:
            data['icon'] = icon
        if button:
            data['button'] = button

        super().__init__(data)


class TextWidget(Widget):
    _key = 'textParagraph'

    def __init__(self, text: str):
        super().__init__(text=text)


class ImageWidget(Widget):
    _key = 'image'

    def __init__(self, image_url: str, on_click: OnClickWidget = None):
        data = dict(imageUrl=image_url)
        if on_click:
            data.update(on_click.as_data())
        super().__init__(data)


@dataclass
class Section:
    header: str = None
    widgets: List[Widget] = field(default_factory=list)

    def add_header(self, text: str):
        self.header = text

    def add_text(
            self,
            content: str,
            title: str = '',
            footer: str = '',
            click_url: str = None,
            icon: str = None,
            button: str = None,
    ):

        widget = KeyValueWidget(
            top=title,
            content=content,
            break_lines=True,
            bottom=footer,
            on_click=OnClickWidget(url=click_url) if click_url else None,
            icon=icon,
            button=button,
        )
        self.widgets.append(widget)

    def add_paragraph(self, text: str):
        self.widgets.append(TextWidget(text=text))

    def add_button(self, url, text: str = None, image_url: str = None, icon: str = None, append: bool = True):
        button = ButtonWidget(url=url, text=text, image_url=image_url, icon=icon)
        if append and self.widgets and 'buttons' in self.widgets[-1]:
            self.widgets[-1]['buttons'].append(button)
        else:
            self.widgets.append(Widget(buttons=[button]))

    def add_image(self, image_url: str, click_url: str = None):
        widget = ImageWidget(
            image_url=image_url,
            on_click=OnClickWidget(url=click_url) if click_url else None,
        )
        self.widgets.append(widget)

    def as_data(self):
        data = {
            'widgets': [widget.as_data() for widget in self.widgets],
        }
        if self.header:
            data['header'] = self.header
        return data

    def __bool__(self):
        return self.header is not None or bool(self.widgets)


@dataclass
class Card:
    header: Widget = None
    sections: List[Section] = field(default_factory=list)

    def add_header(self, title: str, subtitle: str = '', image_url: str = None, style: str = 'IMAGE'):
        self.header = Widget(
            title=title,
            subtitle=subtitle,
            imageUrl=image_url,
            imageStyle=style,
        )

    def add_section(self, section: Section):
        if bool(section):
            self.sections.append(section)

    def as_data(self) -> Dict:
        data = {'sections': [section.as_data() for section in self.sections]}
        if self.header:
            data['header'] = self.header.as_data()
        return data


class ChatsHook:
    def __init__(self, hook_url: str):
        self.hook_url = hook_url

    def _post(self, body: Dict, thread_key: str = None) -> Dict:
        url = self.hook_url
        if thread_key:
            url = f'{url}&threadKey={thread_key}'

        response = requests.post(
            url=url,
            headers={'Content-Type': 'application/json; charset=UTF-8'},
            data=json.dumps(body),
        )
        response.raise_for_status()
        return response.json()

    def send_text(self, text: str, thread_key: str = None) -> Dict:
        body = {'text': text}
        return self._post(body=body, thread_key=thread_key)

    def send_card(self, card: Card, additional_text: str = None, thread_key: str = None) -> Dict:
        body = {
            'cards': [card.as_data()],
        }

        if additional_text:
            body['text'] = additional_text

        return self._post(body=body, thread_key=thread_key)


class ChatsBot(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = ['https://www.googleapis.com/auth/chat.bot']

    def __init__(self, **kwargs):
        super().__init__(
            serviceName='chat',
            version='v1',
            cache_discovery=False,
            **kwargs,
        )

    def _room_path(self, room_id: str) -> str:
        prefix = 'spaces/'
        if not room_id.startswith(prefix):
            room_path = f'{prefix}{room_id}'
        else:
            room_path = room_id
        return room_path

    def _member_path(self, room_id: str, member_id: str) -> str:
        room_path = self._room_path(room_id=room_id)

        prefix = 'members/'
        if not member_id.startswith(prefix):
            member_path = f'{prefix}{member_id}'
        else:
            member_path = member_id

        return f'{room_path}/{member_path}'

    def get_room(self, room_id: str) -> ResourceType:
        return self._execute(
            method=self.client.spaces().get,
            name=self._room_path(room_id=room_id),
        )

    def get_rooms(self) -> Generator[ResourceType, None, None]:
        yield from self._paginate(
            method=self.client.spaces().list,
            result_key='spaces',
        )

    def get_member(self, room_id: str, member_id: str) -> ResourceType:
        name = self._member_path(room_id=room_id, member_id=member_id)
        return self._execute(
            method=self.client.spaces().members().get,
            name=name,
        )

    def get_members(self, room_id: str) -> Generator[ResourceType, None, None]:
        yield from self._paginate(
            method=self.client.spaces().members().list,
            result_key='memberships',
            params={'parent': self._room_path(room_id=room_id)}
        )

    def send_text(self, room_id: str, text: str) -> ResourceType:
        body = {'text': text}

        return self._execute(
            method=self.client.spaces().messages().create,
            parent=self._room_path(room_id=room_id),
            body=body,
        )

    def send_card(self, room_id: str, card: Card, additional_text: str = None) -> ResourceType:
        body = {
            'cards': [card.as_data()],
        }

        if additional_text:
            body['text'] = additional_text

        return self._execute(
            method=self.client.spaces().messages().create,
            parent=self._room_path(room_id=room_id),
            body=body,
        )


__all__ = (
    'Text',
    'Section',
    'Card',
    'ChatsBot',
    'ChatsHook',
)



# More Information <https://cloud.google.com/run/docs/reference/rest>
from typing import Generator

from google.api_core.client_options import ClientOptions
from googleapiclient.discovery import Resource

from gcp_pilot.base import (
    GoogleCloudPilotAPI,
    ResourceType,
    DiscoveryMixin,
)
from gcp_pilot.resource import ServiceAgent


class CloudRun(DiscoveryMixin, GoogleCloudPilotAPI):
    def _service_endpoint(self, location: str = None):
        if not location:
            return 'https://run.googleapis.com'
        return f'https://{location}-run.googleapis.com'

    def _namespace_path(self, project_id: str = None):
        return f'namespaces/{project_id or self.project_id}'

    def _service_path(self, service_name: str, project_id: str = None):
        parent = self._namespace_path(project_id=project_id)
        return f'{parent}/services/{service_name}'

    def _domain_mapping_path(self, domain: str, project_id: str = None):
        parent = self._namespace_path(project_id=project_id)
        return f'{parent}/domainmappings/{domain}'

    def list_services(self, project_id: str = None) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._namespace_path(project_id=project_id),
        )
        yield from self._list(
            method=self.client.namespaces().services().list,
            result_key='items',
            params=params,
        )

    def get_service(self, service_name: str, project_id: str = None, location: str = None) -> ResourceType:
        name = self._service_path(service_name=service_name, project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)
        return self._execute(
            method=client.namespaces().services().get,
            name=name,
        )

    def create_service(
            self,
            service_name: str,
            project_id: str = None,
            location: str = None,
            service_account: str = None,
            trigger_id: str = None,
            image: str = 'gcr.io/cloudrun/placeholder',
            ram: int = 256,
            concurrency: int = 80,
            timeout: int = 300,
            port: int = 8080,
    ) -> ResourceType:
        parent = self._namespace_path(project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)

        service_account = service_account or ServiceAgent.get_compute_service_account(
            project_id=project_id or self.project_id
        )
        labels = {
            'managed-by': 'gcp-cloud-build-deploy-cloud-run',
            'cloud.googleapis.com/location': location,
        }
        if trigger_id:
            labels['gcb-trigger-id'] = trigger_id

        body = {
            'apiVersion': 'serving.knative.dev/v1',
            'kind': 'Service',
            'metadata': {
                'name': service_name,
                'labels': labels,
                'annotations': {
                    'client.knative.dev/user-image': image,
                },
            },
            'spec': {
                'template': {
                    'spec': {
                        'containerConcurrency': concurrency,
                        'timeoutSeconds': timeout,
                        'serviceAccountName': service_account,
                        'containers': [{
                            'image': image,
                            'resources': {'limits': {'cpu': '1000m', 'memory': f'{ram}Mi'}},
                            'ports': [{'containerPort': port}]
                        }]
                    }
                },
            },
        }
        return self._execute(
            method=client.namespaces().services().create,
            parent=parent,
            body=body,
        )

    def list_locations(self, project_id: str = None) -> Generator[ResourceType, None, None]:
        params = dict(
            name=self._project_path(project_id=project_id),
        )
        yield from self._paginate(
            method=self.client.projects().locations().list,
            result_key='locations',
            params=params,
        )

    def list_revisions(self, service_name: str = None, project_id: str = None) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._namespace_path(project_id=project_id),
        )
        items = self._list(
            method=self.client.namespaces().revisions().list,
            result_key='items',
            params=params,
        )
        for item in items:
            if not service_name or item['metadata']['labels']['serving.knative.dev/service'] == service_name:
                yield item

    def list_domain_mappings(self, project_id: str = None) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._namespace_path(project_id=project_id),
        )
        yield from self._list(
            method=self.client.namespaces().domainmappings().list,
            result_key='items',
            params=params,
        )

    def get_domain_mapping(self, domain: str, project_id: str = None, location: str = None) -> ResourceType:
        name = self._domain_mapping_path(domain=domain, project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)
        return self._execute(
            method=client.namespaces().domainmappings().get,
            name=name,
        )

    def delete_domain_mapping(self, domain: str, project_id: str = None, location: str = None) -> ResourceType:
        name = self._domain_mapping_path(domain=domain, project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)
        return self._execute(
            method=client.namespaces().domainmappings().delete,
            name=name,
        )

    def create_domain_mapping(
            self,
            domain: str,
            service_name: str,
            project_id: str = None,
            location: str = None,
    ) -> ResourceType:
        parent = self._namespace_path(project_id=project_id)
        client = self._get_localized_client(project_id=project_id, location=location)

        body = {
            'apiVersion': 'domains.cloudrun.com/v1',
            'kind': 'DomainMapping',
            'metadata': {'name': domain},
            'spec': {'routeName': service_name, 'certificateMode': 'AUTOMATIC'},
        }
        return self._execute(
            method=client.namespaces().domainmappings().create,
            parent=parent,
            body=body,
        )

    def _get_localized_client(self, project_id: str = None, location: str = None) -> Resource:
        # Reminder: List methods do not require a localized client
        if not location:
            if not project_id:
                location = self.location
            else:
                location = self._get_project_default_location(project_id=project_id)
        return self._build_client(location=location)

    def _build_client(self, location: str = None, **kwargs) -> Resource:
        options = ClientOptions(api_endpoint=self._service_endpoint(location=location))
        kwargs.update(dict(
            serviceName='run',
            version='v1',
            cache_discovery=False,
            client_options=options,
        ))
        return super()._build_client(**kwargs)


__all__ = (
    'CloudRun',
)



# More Information: <https://cloud.google.com/resource-manager/reference/rest>

from typing import List, Tuple, Generator

from gcp_pilot import exceptions
from gcp_pilot.base import AccountManagerMixin, GoogleCloudPilotAPI, PolicyType, DiscoveryMixin, ResourceType


class ResourceManager(AccountManagerMixin, DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self, **kwargs):
        super().__init__(
            serviceName='cloudresourcemanager',
            version='v1',
            cache_discovery=False,
            **kwargs,
        )

    def get_policy(self, project_id: str = None, version: int = 1) -> PolicyType:
        return self._execute(
            method=self.client.projects().getIamPolicy,
            resource=project_id or self.project_id,
            body={"options": {"requestedPolicyVersion": version}},
        )

    def set_policy(self, policy: PolicyType, project_id: str = None) -> PolicyType:
        if not policy['bindings']:
            raise exceptions.NotAllowed("Too dangerous to set policy with empty bindings")

        return self._execute(
            method=self.client.projects().setIamPolicy,
            resource=project_id or self.project_id,
            body={"policy": policy, "updateMask": 'bindings'},
        )

    async def add_member(self, email: str, role: str, project_id: str = None) -> PolicyType:
        policy = self.get_policy(project_id=project_id)
        changed_policy = self._bind_email_to_policy(email=email, role=role, policy=policy)
        return self.set_policy(policy=changed_policy, project_id=project_id)

    async def remove_member(self, email: str, role: str, project_id: str = None) -> PolicyType:
        policy = self.get_policy(project_id=project_id)
        changed_policy = self._unbind_email_from_policy(email=email, role=role, policy=policy)
        return self.set_policy(policy=changed_policy, project_id=project_id)

    def get_project(self, project_id: str) -> ResourceType:
        return self._execute(
            method=self.client.projects().get,
            projectId=project_id or self.project_id,
        )

    async def allow_impersonation(self, email: str, project_id: str = None) -> PolicyType:
        return await self.add_member(
            email=email,
            role='roles/iam.serviceAccountTokenCreator',
            project_id=project_id,
        )


class ServiceAgent:
    agents = {
        'AI Platform Custom Code Service Agent': ('gcp-sa-aiplatform-cc.iam.gserviceaccount.com', None),
        'AI Platform Service Agent': ('gcp-sa-aiplatform.iam.gserviceaccount.com', None),
        'ASM Mesh Data Plane Service Account': (
            'gcp-sa-meshdataplane.iam.gserviceaccount.com', 'roles/meshdataplane.serviceAgent'),
        'Anthos Audit Service Account': ('gcp-sa-anthosaudit.iam.gserviceaccount.com', None),
        'Anthos Config Management Service Account': (
            'gcp-sa-anthosconfigmanagement.iam.gserviceaccount.com', 'roles/anthosconfigmanagement.serviceAgent'),
        'Anthos Identity Service Account': (
            'gcp-sa-anthosidentityservice.iam.gserviceaccount.com', 'roles/anthosidentityservice.serviceAgent'),
        'Anthos Service Account': ('gcp-sa-anthos.iam.gserviceaccount.com', 'roles/anthos.serviceAgent'),
        'Anthos Service Mesh Service Account': ('gcp-sa-servicemesh.iam.gserviceaccount.com', None),
        'Apigee Service Agent': ('gcp-sa-apigee.iam.gserviceaccount.com', 'roles/apigee.serviceAgent'),
        'App Engine Flexible Environment Service Agent': (
            'gae-api-prod.google.com.iam.gserviceaccount.com', 'roles/appengineflex.serviceAgent'),
        'Artifact Registry Service Agent': (
            'gcp-sa-artifactregistry.iam.gserviceaccount.com', 'roles/artifactregistry.serviceAgent'),
        'AssuredWorkloads Service Account': ('gcp-sa-assuredworkloads.iam.gserviceaccount.com', None),
        'AutoML Recommendations Service Account': (
            'gcp-sa-recommendationengine.iam.gserviceaccount.com', 'roles/automlrecommendations.serviceAgent'),
        'AutoML Service Agent': ('gcp-sa-automl.iam.gserviceaccount.com', 'roles/automl.serviceAgent'),
        'BigQuery Connection Service Agent': (
            'gcp-sa-bigqueryconnection.iam.gserviceaccount.com', 'roles/bigqueryconnection.serviceAgent'),
        'BigQuery Data Transfer Service Agent': (
            'gcp-sa-bigquerydatatransfer.iam.gserviceaccount.com', 'roles/bigquerydatatransfer.serviceAgent'),
        'BigQuery Omni Service Agent': ('gcp-sa-prod-bigqueryomni.iam.gserviceaccount.com', None),
        'Binary Authorization Service Agent': (
            'gcp-sa-binaryauthorization.iam.gserviceaccount.com', 'roles/binaryauthorization.serviceAgent'),
        'Cloud AI Platform Notebooks Service Account': (
            'gcp-sa-notebooks.iam.gserviceaccount.com', 'roles/notebooks.serviceAgent'),
        'Cloud API Gateway Management Plane Service Account': (
            'gcp-sa-apigateway-mgmt.iam.gserviceaccount.com', None),
        'Cloud API Gateway Service Account': ('gcp-sa-apigateway.iam.gserviceaccount.com', None),
        'Cloud Asset Service Agent': (
            'gcp-sa-cloudasset.iam.gserviceaccount.com', 'roles/cloudasset.serviceAgent'),
        'Cloud Bigtable Service Agent': ('gcp-sa-bigtable.iam.gserviceaccount.com', None),
        'Cloud Build Service Agent': (
            'gcp-sa-cloudbuild.iam.gserviceaccount.com', 'roles/cloudbuild.serviceAgent'),
        'Cloud Composer Service Agent': (
            'cloudcomposer-accounts.iam.gserviceaccount.com', 'roles/composer.serviceAgent'),
        'Cloud Data Fusion Service Account': (
            'gcp-sa-datafusion.iam.gserviceaccount.com', 'roles/datafusion.serviceAgent'),
        'Cloud Database Migration Service Account': ('gcp-sa-datamigration.iam.gserviceaccount.com', None),
        'Cloud Dataflow Service Account': (
            'dataflow-service-producer-prod.iam.gserviceaccount.com', 'roles/dataflow.serviceAgent'),
        'Cloud Endpoints Service Agent': (
            'gcp-sa-endpoints.iam.gserviceaccount.com', 'roles/endpoints.serviceAgent'),
        'Cloud File Storage Service Account': ('cloud-filer.iam.gserviceaccount.com', 'roles/file.serviceAgent'),
        'Cloud Healthcare Service Agent': (
            'gcp-sa-healthcare.iam.gserviceaccount.com', 'roles/healthcare.serviceAgent'),
        'Cloud IoT Core Service Agent': (
            'gcp-sa-cloudiot.iam.gserviceaccount.com', 'roles/cloudiot.serviceAgent'),
        'Cloud KMS Service Agent': ('gcp-sa-cloudkms.iam.gserviceaccount.com', 'roles/cloudkms.serviceAgent'),
        'Cloud Life Sciences Service Agent': (
            'gcp-sa-lifesciences.iam.gserviceaccount.com', 'roles/lifesciences.serviceAgent'),
        'Cloud Logging Service Account': ('gcp-sa-logging.iam.gserviceaccount.com', None),
        'Cloud Managed Identities Service Agent': (
            'gcp-sa-mi.iam.gserviceaccount.com', 'roles/managedidentities.serviceAgent'),
        'Cloud Memorystore Memcache Service Agent': (
            'cloud-memcache-sa.iam.gserviceaccount.com', 'roles/memcache.serviceAgent'),
        'Cloud Memorystore Redis Service Agent': (
            'cloud-redis.iam.gserviceaccount.com', 'roles/redis.serviceAgent'),
        'Cloud Network Management Service Account': (
            'gcp-sa-networkmanagement.iam.gserviceaccount.com', 'roles/networkmanagement.serviceAgent'),
        'Cloud Pub/Sub Service Account': ('gcp-sa-pubsub.iam.gserviceaccount.com', None),
        'Cloud SQL Service Account': ('gcp-sa-cloud-sql.iam.gserviceaccount.com', 'roles/cloudsql.serviceAgent'),
        'Cloud Scheduler Service Account': (
            'gcp-sa-cloudscheduler.iam.gserviceaccount.com', 'roles/cloudscheduler.serviceAgent'),
        'Cloud Security Command Center Notification Service Account': (
            'gcp-sa-scc-notification.iam.gserviceaccount.com', 'roles/securitycenter.notificationServiceAgent'),
        'Cloud Source Repositories Service Agent': (
            'sourcerepo-service-accounts.iam.gserviceaccount.com', 'roles/sourcerepo.serviceAgent'),
        'Cloud Spanner Production Service Account': ('gcp-sa-spanner.iam.gserviceaccount.com', None),
        'Cloud Storage for Firebase Service Agent': (
            'gcp-sa-firebasestorage.iam.gserviceaccount.com', 'roles/firebasestorage.serviceAgent'),
        'Cloud Tasks Service Account': (
            'gcp-sa-cloudtasks.iam.gserviceaccount.com', 'roles/cloudtasks.serviceAgent'),
        'Cloud Trace Service Account': ('gcp-sa-cloud-trace.iam.gserviceaccount.com', None),
        'Cloud Translation Service Agent': (
            'gcp-sa-translation.iam.gserviceaccount.com', 'roles/cloudtranslate.serviceAgent'),
        'Cloud VM Migration Service Account': ('gcp-sa-vmmigration.iam.gserviceaccount.com', None),
        'Cloud Web Security Scanner Service Agent': (
            'gcp-sa-websecurityscanner.iam.gserviceaccount.com', 'roles/websecurityscanner.serviceAgent'),
        'Cloud Workflows Service Agent': (
            'gcp-sa-workflows.iam.gserviceaccount.com', 'roles/workflows.serviceAgent'),
        'Compute Engine Service Agent': ('compute-system.iam.gserviceaccount.com', 'roles/compute.serviceAgent'),
        'Compute Scanning Service Agent': (
            'gcp-sa-computescanning.iam.gserviceaccount.com', 'roles/computescanning.serviceAgent'),
        'Container Analysis Service Agent': (
            'container-analysis.iam.gserviceaccount.com', 'roles/containeranalysis.ServiceAgent'),
        'Container Scanning Service Agent': (
            'gcp-sa-containerscanning.iam.gserviceaccount.com', 'roles/containerscanning.ServiceAgent'),
        'Container Threat Detection Service Agent': (
            'gcp-sa-ktd-control.iam.gserviceaccount.com', 'roles/containerthreatdetection.serviceAgent'),
        'Data Labeling Service Account': (
            'gcp-sa-datalabeling.iam.gserviceaccount.com', 'roles/datalabeling.serviceAgent'),
        'Data Studio Service Account': (
            'gcp-sa-datastudio.iam.gserviceaccount.com', 'roles/datastudio.serviceAgent'),
        'Dataproc Metastore Service Account': ('gcp-sa-metastore.iam.gserviceaccount.com', None),
        'Dialogflow Service Agent': (
            'gcp-sa-dialogflow.iam.gserviceaccount.com', 'roles/dialogflow.serviceAgent'),
        'DocumentAI Core Service Agent': (
            'gcp-sa-prod-dai-core.iam.gserviceaccount.com', 'roles/documentaicore.serviceAgent'),
        'Endpoints Consumer Portal Service Agent': (
            'endpoints-portal.iam.gserviceaccount.com', 'roles/endpointsportal.serviceAgent'),
        'Eventarc Service Agent': ('gcp-sa-eventarc.iam.gserviceaccount.com', 'roles/eventarc.serviceAgent'),
        'External Key Management Service Service Account': ('gcp-sa-ekms.iam.gserviceaccount.com', None),
        'Firebase Extensions Service Agent': (
            'gcp-sa-firebasemods.iam.gserviceaccount.com', 'roles/firebasemods.serviceAgent'),
        'Firebase Rules Service Agent': ('firebase-rules.iam.gserviceaccount.com', None),
        'Firewall Insights Service Account': (
            'gcp-sa-firewallinsights.iam.gserviceaccount.com', 'roles/firewallinsights.serviceAgent'),
        'GKE Hub API Service Account': ('gcp-sa-gkehub.iam.gserviceaccount.com', 'roles/gkehub.serviceAgent'),
        'Game Services Agent': ('gcp-sa-gameservices.iam.gserviceaccount.com', 'roles/gameservices.serviceAgent'),
        'Google Cloud Dataproc Service Agent': (
            'dataproc-accounts.iam.gserviceaccount.com', 'roles/dataproc.serviceAgent'),
        'Google Cloud Functions Service Agent': (
            'gcf-admin-robot.iam.gserviceaccount.com', 'roles/cloudfunctions.serviceAgent'),
        'Google Cloud ML Engine Service Agent': (
            'cloud-ml.google.com.iam.gserviceaccount.com', 'roles/ml.serviceAgent'),
        'Google Cloud OS Config Service Agent': (
            'gcp-sa-osconfig.iam.gserviceaccount.com', 'roles/osconfig.serviceAgent'),
        'Google Cloud Run Service Agent': (
            'serverless-robot-prod.iam.gserviceaccount.com', 'roles/run.serviceAgent'),
        'Google Container Registry Service Agent': (
            'containerregistry.iam.gserviceaccount.com', 'roles/containerregistry.ServiceAgent'),
        'Google Genomics Service Agent': (
            'genomics-api.google.com.iam.gserviceaccount.com', 'roles/genomics.serviceAgent'),
        'Kubernetes Engine Service Agent': (
            'container-engine-robot.iam.gserviceaccount.com', 'roles/container.serviceAgent'),
        'Mesh Config Service Account': (
            'gcp-sa-meshconfig.iam.gserviceaccount.com', 'roles/meshconfig.serviceAgent'),
        'Monitoring Notification Service Account': (
            'gcp-sa-monitoring-notification.iam.gserviceaccount.com', 'roles/monitoring.notificationServiceAgent'),
        'Multi Cluster Ingress Service Account': (
            'gcp-sa-multiclusteringress.iam.gserviceaccount.com', 'roles/multiclusteringress.serviceAgent'),
        'Multi cluster metering Service Account': (
            'gcp-sa-mcmetering.iam.gserviceaccount.com', 'roles/multiclustermetering.serviceAgent'),
        'Private CA Service Account': ('gcp-sa-privateca.iam.gserviceaccount.com', None),
        'Remote Build Execution Service Agent': (
            'remotebuildexecution.iam.gserviceaccount.com', 'roles/remotebuildexecution.serviceAgent'),
        'Retail Service Account': ('gcp-sa-retail.iam.gserviceaccount.com', 'roles/retail.serviceAgent'),
        'Secret Manager Service Account': ('gcp-sa-secretmanager.iam.gserviceaccount.com', None),
        'Serverless VPC Access Service Agent': (
            'gcp-sa-vpcaccess.iam.gserviceaccount.com', 'roles/vpcaccess.serviceAgent'),
        'Service Consumer Management Service Agent': (
            'service-consumer-management.iam.gserviceaccount.com', None),
        'Service Directory Service Account': ('gcp-sa-servicedirectory.iam.gserviceaccount.com', None),
        'Service Networking Service Agent': (
            'service-networking.iam.gserviceaccount.com', 'roles/servicenetworking.serviceAgent'),
        'TPU Service Agent': ('cloud-tpu.iam.gserviceaccount.com', 'roles/tpu.serviceAgent'),
        'TPU Service Agent (v2)': ('gcp-sa-tpu.iam.gserviceaccount.com', 'roles/cloudtpu.serviceAgent'),
        'Transcoder Service Account': ('gcp-sa-transcoder.iam.gserviceaccount.com', None)
    }

    _suffixes = ['Service Agent', 'Agent', 'Service Account']

    @classmethod
    def get_available_agents(cls) -> Generator[str, None, None]:
        for agent in cls.agents:
            name = agent
            for suffix in cls._suffixes:
                name = name.replace(suffix, '')
            yield name

    @classmethod
    def _load_from_tsv(cls, filepath: str):
        # Load this <https://cloud.google.com/iam/docs/service-agents> table
        data = {}
        for r in open(filepath, 'r').readlines()[1:]:
            name, domain, role = r.strip().split('\t')
            if 'roles/' not in role:
                role_name = None
            else:
                role_name = role.replace('(', '').replace(')', '')
            data[name] = (domain, role_name)
        return data

    @classmethod
    def _find(cls, service_name: str) -> Tuple[str, str]:
        candidates = [f'{service_name} {suffix}' for suffix in cls._suffixes]

        for candidate in candidates:
            try:
                return cls.agents[candidate]
            except KeyError:
                continue

        raise exceptions.NotFound(f"Service {service_name} not found")

    @classmethod
    def get_email(cls, service_name: str, project_id: str) -> str:
        domain, _ = cls._find(service_name=service_name)
        project_number = cls.get_project_number(project_id=project_id)
        return f'service-{project_number}@{domain}'

    @classmethod
    def get_role(cls, service_name: str) -> str:
        _, role = cls._find(service_name=service_name)
        return role

    @classmethod
    def get_project_number(cls, project_id: str) -> int:
        # TODO: cache this
        project = ResourceManager().get_project(project_id=project_id)
        return project['projectNumber']

    @classmethod
    async def restore(cls, services: List[str], project_id: str) -> None:
        rm = ResourceManager()
        for service_name in services:
            email = ServiceAgent.get_email(service_name=service_name, project_id=project_id)
            role = ServiceAgent.get_role(service_name=service_name)
            if role:
                try:
                    await rm.add_member(email=email, role=role)
                    print(f'[O] {service_name}')
                except exceptions.ValidationError:
                    print(f'[X] {service_name}')

    @classmethod
    def get_compute_service_account(cls, project_id: str) -> str:
        project_number = cls.get_project_number(project_id=project_id)
        return f'{project_number}-compute@developer.gserviceaccount.com'

    @classmethod
    def get_cloud_build_service_account(cls, project_id: str = None) -> str:
        project_number = cls.get_project_number(project_id=project_id)
        return f'{project_number}@cloudbuild.gserviceaccount.com'


__all__ = (
    'ResourceManager',
    'ServiceAgent',
)
